@@ -2,11 +2,11 @@
     flask_security
     ~~~~~~~~~~~~~~
 
-    Flask-Security is a Flask extension that aims to add quick and simple
-    security via Flask-Login, Flask-Principal, Flask-WTF, and passlib.
+    Flask-Security is a Flask extension that aims to add comprehensive security
+    to Flask applications.
 
     :copyright: (c) 2012-2019 by Matt Wright.
-    :copyright: (c) 2019-2023 by J. Christopher Wagner.
+    :copyright: (c) 2019-2024 by J. Christopher Wagner.
     :license: MIT, see LICENSE for more details.
 """
 
@@ -17,7 +17,6 @@ from .core import (
     RoleMixin,
     UserMixin,
     WebAuthnMixin,
-    AnonymousUser,
     FormInfo,
     current_user,
 )
@@ -61,6 +60,7 @@ from .forms import (
 )
 from .mail_util import MailUtil
 from .oauth_glue import OAuthGlue
+from .oauth_provider import FsOAuthProvider
 from .password_util import PasswordUtil
 from .phone_util import PhoneUtil
 from .recovery_codes import (
@@ -79,6 +79,7 @@ from .signals import (
     tf_security_token_sent,
     tf_disabled,
     user_authenticated,
+    user_unauthenticated,
     user_confirmed,
     user_registered,
     user_not_registered,
@@ -111,6 +112,7 @@ from .utils import (
     login_user,
     logout_user,
     lookup_identity,
+    naive_utcnow,
     password_breached_validator,
     password_complexity_validator,
     password_length_validator,
@@ -134,4 +136,4 @@ from .webauthn import (
 )
 from .webauthn_util import WebauthnUtil
 
-__version__ = "5.1.2"
+__version__ = "5.4.1"
