@@ -5,19 +5,21 @@
     Flask-Security utils module
 
     :copyright: (c) 2012-2019 by Matt Wright.
-    :copyright: (c) 2019-2022 by J. Christopher Wagner (jwag).
+    :copyright: (c) 2019-2024 by J. Christopher Wagner (jwag).
     :license: MIT, see LICENSE for more details.
 """
+
+from __future__ import annotations
+
 import abc
 import base64
-import datetime
+from datetime import datetime, timedelta, timezone
 from functools import partial
 import hashlib
 import hmac
-from pkg_resources import parse_version
 import time
 import typing as t
-from urllib.parse import parse_qsl, parse_qs, urlsplit, urlunsplit, urlencode
+from urllib.parse import parse_qsl, quote, urlsplit, urlunsplit, urlencode
 import urllib.request
 import urllib.error
 import warnings
@@ -37,11 +39,11 @@ from flask_login import logout_user as _logout_user
 from flask_login import current_user
 from flask_login import COOKIE_NAME as REMEMBER_COOKIE_NAME
 from flask_principal import AnonymousIdentity, Identity, identity_changed, Need
-from flask_wtf import csrf
+from flask_wtf import csrf, FlaskForm
 from wtforms import ValidationError
 from itsdangerous import BadSignature, SignatureExpired
-from werkzeug import __version__ as werkzeug_version
 from werkzeug.local import LocalProxy
+from werkzeug.datastructures import MultiDict
 
 from .quart_compat import best, get_quart_status
 from .proxies import _security, _datastore, _pwd_context, _hashing_context
@@ -50,12 +52,8 @@ from .signals import user_authenticated
 if t.TYPE_CHECKING:  # pragma: no cover
     from flask import Flask, Response
     from flask.typing import ResponseValue
-    from flask_wtf import FlaskForm
     from .datastore import User
 
-SB = t.Union[str, bytes]
-
-
 localize_callback = LocalProxy(lambda: _security.i18n_domain.gettext)
 
 FsPermNeed = partial(Need, "fsperm")
@@ -106,6 +104,25 @@ else:
         return response
 
 
+# From a miguel grinberg blog around dealing with 3.12.
+# Our default SQLAlchemy Datetime is naive.
+# Note that most code should call _security.datetime_factory()
+def aware_utcnow():
+    return datetime.now(timezone.utc)
+
+
+def aware_utcfromtimestamp(timestamp):
+    return datetime.fromtimestamp(timestamp, timezone.utc)
+
+
+def naive_utcnow():
+    return aware_utcnow().replace(tzinfo=None)
+
+
+def naive_utcfromtimestamp(timestamp):
+    return aware_utcfromtimestamp(timestamp).replace(tzinfo=None)
+
+
 def find_csrf_field_name():
     """
     We need to clear it on logout (since that isn't being done by Flask-WTF).
@@ -121,19 +138,38 @@ def find_csrf_field_name():
     return None
 
 
+def is_user_authenticated(user: User | None) -> bool:
+    """
+    return True is user is authenticated.
+
+    With Flask-Login <=0.6.x and Flask-Security <5.4 current_user was always
+    set - for non-authenticated users it pointed to an AnonymousUser
+    Flask-Login is experimenting (11/5/23) with a LOGIN_NO_ANONYMOUS which will set
+    current_user to None and deprecate is_authenticated (current_user non None implies
+    authenticated).
+    We have a configuration variable ANONYMOUS_USER_DISABLED which if true will force
+    current_user to None on unauthenticated as well
+    """
+    if config_value("ANONYMOUS_USER_DISABLED"):
+        # Note that user often is current_user which is a proxy and isn't ever actually
+        # 'None'
+        return bool(user)
+    return bool(user and user.is_authenticated)
+
+
 def login_user(
-    user: "User",
-    remember: t.Optional[bool] = None,
-    authn_via: t.Optional[t.List[str]] = None,
+    user: User,
+    remember: bool | None = None,
+    authn_via: list[str] | None = None,
 ) -> bool:
     """Perform the login routine.
 
-    If *SECURITY_TRACKABLE* is used, make sure you commit changes after this
+    If :py:data:`SECURITY_TRACKABLE` is used, make sure you commit changes after this
     request (i.e. ``app.security.datastore.commit()``).
 
     :param user: The user to login
     :param remember: Flag specifying if the remember cookie should be set.
-                     If ``None`` use value of SECURITY_DEFAULT_REMEMBER_ME
+                     If ``None`` use value of :py:data:`SECURITY_DEFAULT_REMEMBER_ME`
     :param authn_via: A list of strings denoting which mechanism(s) the user
         authenticated with.
         These should be one or more of ["password", "sms", "authenticator", "email"] or
@@ -168,12 +204,14 @@ def login_user(
     session["fs_paa"] = time.time()  # Primary authentication at - timestamp
 
     identity_changed.send(
-        current_app._get_current_object(),  # type: ignore
+        current_app._get_current_object(),  # type: ignore[attr-defined]
+        _async_wrapper=current_app.ensure_sync,
         identity=Identity(user.fs_uniquifier),
     )
 
     user_authenticated.send(
-        current_app._get_current_object(),  # type: ignore
+        current_app._get_current_object(),  # type: ignore[attr-defined]
+        _async_wrapper=current_app.ensure_sync,
         user=user,
         authn_via=authn_via,
     )
@@ -189,7 +227,13 @@ def logout_user() -> None:
     identity is now the `AnonymousIdentity`
     """
 
-    for key in ("identity.name", "identity.auth_type", "fs_paa", "fs_gexp"):
+    for key in (
+        "identity.name",
+        "identity.auth_type",
+        "fs_paa",
+        "fs_gexp",
+        "fs_oauth_next",
+    ):
         session.pop(key, None)
 
     # Clear csrf token between sessions.
@@ -203,15 +247,17 @@ def logout_user() -> None:
         g.pop(csrf_field_name, None)
     session["fs_cc"] = "clear"
     identity_changed.send(
-        current_app._get_current_object(), identity=AnonymousIdentity()  # type: ignore
+        current_app._get_current_object(),  # type: ignore
+        _async_wrapper=current_app.ensure_sync,
+        identity=AnonymousIdentity(),
     )
     _logout_user()
 
 
 def check_and_update_authn_fresh(
-    within: datetime.timedelta,
-    grace: datetime.timedelta,
-    method: t.Optional[str] = None,
+    within: timedelta,
+    grace: timedelta,
+    method: str | None = None,
 ) -> bool:
     """Check if user authenticated within specified time and update grace period.
 
@@ -257,12 +303,11 @@ def check_and_update_authn_fresh(
         # No session, you can't play.
         return False
 
-    now = datetime.datetime.utcnow()
+    now = naive_utcnow()
     new_exp = now + grace
     grace_ts = int(new_exp.timestamp())
 
-    fs_gexp = session.get("fs_gexp", None)
-    if fs_gexp:
+    if fs_gexp := session.get("fs_gexp", None):
         if now.timestamp() < fs_gexp:
             # Within grace period - extend it, and we're good.
             session["fs_gexp"] = grace_ts
@@ -273,7 +318,7 @@ def check_and_update_authn_fresh(
         session["fs_gexp"] = grace_ts
         return False
 
-    authn_time = datetime.datetime.utcfromtimestamp(session["fs_paa"])
+    authn_time = naive_utcfromtimestamp(session["fs_paa"])
     # allow for some time drift where it's possible authn_time is in the future
     # but let's be cautious and not allow arbitrary future times
     delta = now - authn_time
@@ -283,15 +328,13 @@ def check_and_update_authn_fresh(
     return False
 
 
-def get_hmac(password: SB) -> bytes:
+def get_hmac(password: str | bytes) -> bytes:
     """Returns a Base64 encoded HMAC+SHA512 of the password signed with
-    the salt specified by *SECURITY_PASSWORD_SALT*.
+    the salt specified by :py:data:`SECURITY_PASSWORD_SALT`.
 
     :param password: The password to sign
     """
-    salt = config_value("PASSWORD_SALT")
-
-    if salt is None:
+    if not (salt := config_value("PASSWORD_SALT")):
         raise RuntimeError(
             "The configuration value `SECURITY_PASSWORD_SALT` must "
             "not be None when the value of `SECURITY_PASSWORD_HASH` is "
@@ -302,7 +345,7 @@ def get_hmac(password: SB) -> bytes:
     return base64.b64encode(h.digest())
 
 
-def verify_password(password: SB, password_hash: SB) -> bool:
+def verify_password(password: str | bytes, password_hash: str | bytes) -> bool:
     """Returns ``True`` if the password matches the supplied hash.
 
     :param password: A plaintext password to verify
@@ -318,7 +361,7 @@ def verify_password(password: SB, password_hash: SB) -> bool:
     return _pwd_context.verify(password, password_hash)
 
 
-def verify_and_update_password(password: SB, user: "User") -> bool:
+def verify_and_update_password(password: str | bytes, user: User) -> bool:
     """Returns ``True`` if the password is valid for the specified user.
 
     Additionally, the hashed password in the database is updated if the
@@ -349,13 +392,14 @@ def verify_and_update_password(password: SB, user: "User") -> bool:
     return verified
 
 
-def hash_password(password: SB) -> t.Any:
+def hash_password(password: str | bytes) -> str:
     """Hash the specified plaintext password.
 
-    Unless the hash algorithm (as specified by `SECURITY_PASSWORD_HASH`) is listed in
-    the configuration variable `SECURITY_PASSWORD_SINGLE_HASH`,
+    Unless the hash algorithm (as specified by
+    :py:data:`SECURITY_PASSWORD_HASH`) is listed in
+    the configuration variable :py:data:`SECURITY_PASSWORD_SINGLE_HASH`,
     perform a double hash - first create an HMAC from the plaintext password
-    and the value of `SECURITY_PASSWORD_SALT`,
+    and the value of :py:data:`SECURITY_PASSWORD_SALT`,
     then use the configured hashing algorithm.
     This satisfies OWASP/ASVS section 2.4.5: 'provide additional
     iteration of a key derivation'.
@@ -405,9 +449,8 @@ def suppress_form_csrf():
     if get_request_attr("fs_ignore_csrf"):
         # This is the case where CsrfProtect was already called (e.g. @auth_required)
         return {"csrf": False}
-    if (
-        config_value("CSRF_IGNORE_UNAUTH_ENDPOINTS")
-        and not current_user.is_authenticated
+    if config_value("CSRF_IGNORE_UNAUTH_ENDPOINTS") and not is_user_authenticated(
+        current_user
     ):
         return {"csrf": False}
     return {}
@@ -424,10 +467,54 @@ def do_flash(message: str, category: str) -> None:
         flash(message, category)
 
 
-def get_url(endpoint_or_url: str, qparams: t.Optional[t.Dict[str, str]] = None) -> str:
+def parse_auth_token(auth_token: str) -> dict[str, t.Any]:
+    """Parse an authentication token.
+    This will raise an exception if not properly signed or expired
+    """
+    tdata = dict()
+
+    # This can raise BadSignature or SignatureExpired exceptions from itsdangerous
+    raw_data = _security.remember_token_serializer.loads(
+        auth_token, max_age=config_value("TOKEN_MAX_AGE")
+    )
+
+    # Version 3.x generated tokens that map to data with 3 elements,
+    # and fs_uniquifier was on last element.
+    # Version 4.0.0 generates tokens that map to data with only 1 element,
+    # which maps to fs_uniquifier.
+    # Version 5 and up are already a dict (with a version #)
+    if isinstance(raw_data, dict):
+        # new format - starting at ver=5
+        if not all(k in raw_data for k in ["ver", "uid", "exp", "sid"]):
+            raise ValueError("Token missing keys")
+        tdata = raw_data
+        if ts := tdata.get("exp"):
+            if ts < int(time.time()):
+                raise SignatureExpired("token[exp] value expired")
+    else:
+        # old tokens that were lists
+        if len(raw_data) == 1:
+            # version 4
+            tdata["ver"] = "4"
+            tdata["uid"] = raw_data[0]
+        else:
+            # version 3
+            tdata["ver"] = "3"
+            tdata["uid"] = raw_data[2]
+
+    return tdata
+
+
+def get_url(endpoint_or_url: str, qparams: dict[str, str] | None = None) -> str:
     """Returns a URL if a valid endpoint is found. Otherwise, returns the
     provided value.
 
+    .. warning::
+        If an endpoint ISN'T provided, then it is assumed that the URL
+        is external to Flask and if the spa configuration REDIRECT_HOST
+        is set will redirect to that host. This could be an issue in
+        development.
+
     :param endpoint_or_url: The endpoint name or URL to default to
     :param qparams: additional query params to add to end of url
     :return: URL
@@ -438,7 +525,6 @@ def get_url(endpoint_or_url: str, qparams: t.Optional[t.Dict[str, str]] = None)
         # This is an external URL (no endpoint defined in app)
         # For (mostly) testing - allow changing/adding the url - for example
         # add a different host:port for cases where the UI is running
-        # add a different host:port for cases where the UI is running
         # separately.
         if config_value("REDIRECT_HOST"):
             url = transform_url(
@@ -458,7 +544,7 @@ def slash_url_suffix(url, suffix):
 
 
 def transform_url(
-    url: str, qparams: t.Optional[t.Dict[str, str]] = None, **kwargs: str
+    url: str, qparams: dict[str, str] | None = None, **kwargs: str
 ) -> str:
     """Modify url
 
@@ -478,7 +564,7 @@ def transform_url(
 
 
 def get_security_endpoint_name(endpoint):
-    return f"{_security.blueprint_name}.{endpoint}"
+    return f"{config_value('BLUEPRINT_NAME')}.{endpoint}"
 
 
 def url_for_security(endpoint: str, **values: t.Any) -> str:
@@ -543,93 +629,101 @@ def validate_redirect_url(url: str) -> bool:
             return True
         else:
             return False
-    if config_value("REDIRECT_VALIDATE_MODE") == "regex":
-        matcher = _security._redirect_validate_re.match(url)
-        return matcher is None
     return True
 
 
-def get_post_action_redirect(config_key: str, declared: t.Optional[str] = None) -> str:
-    # All this nonsense due to mypy
-    arg_next_url = None
-    arg_next = request.args.get("next", None)
-    if arg_next:
-        arg_next_url = get_url(arg_next)
-    form_next_url = None
-    form_next = request.form.get("next", None)
-    if form_next:
-        form_next_url = get_url(form_next)
+def get_post_action_redirect(
+    config_key: str, next_loc: FlaskForm | MultiDict | dict | None
+) -> str:
+    """
+    There is a security angle here - the result of this method is
+    sent to Flask::redirect() - and we need to be sure that it can't be
+    interpreted as a user-input external URL - that would mean we would
+    have an 'open-redirect' vulnerability.
+    """
+    rurl = propagate_next(find_redirect(config_key), next_loc)
+    (scheme, netloc, path, query, fragment) = urlsplit(rurl)
+    safe_url = urlunsplit((scheme, netloc, quote(path), query, fragment))
+    return safe_url
+
+
+def get_post_login_redirect() -> str:
+    return get_post_action_redirect("SECURITY_POST_LOGIN_VIEW", request.form)
+
+
+def get_post_register_redirect() -> str:
+    return get_post_action_redirect("SECURITY_POST_REGISTER_VIEW", request.form)
+
+
+def get_post_logout_redirect() -> str:
+    return get_post_action_redirect("SECURITY_POST_LOGOUT_VIEW", request.form)
+
+
+def get_post_verify_redirect() -> str:
+    return get_post_action_redirect("SECURITY_POST_VERIFY_VIEW", request.form)
+
+
+def find_redirect(key: str) -> str:
+    """Returns the URL to redirect to.
+
+    :param key: The  application configuration key to search for
+    """
+    app_url = None
+    if app_value := current_app.config[key.upper()]:
+        app_url = get_url(app_value)
+    rv = app_url or str(current_app.config.get("APPLICATION_ROOT", "/"))
+    return rv
+
+
+def propagate_next(fallback_url: str, form: FlaskForm | MultiDict | dict | None) -> str:
+    """Compute appropriate redirect URL
+    The application can add a 'next' query parameter or have 'next' as a form field.
+    If either exist, make sure they are valid (not pointing to external location)
+    If neither, return the fallback_url
+
+    Can be passed either request.form
+     (which is really a MultiDict OR a real form OR a dict with a 'next' key).
+    """
+    form_next = None
+    if form and isinstance(form, FlaskForm):
+        if hasattr(form, "next") and form.next.data:
+            form_next = form.next.data
+    elif form and form.get("next", None):
+        form_next = str(form.get("next"))
+    arg_next = request.args.get("next")
+
     urls = [
-        arg_next_url,
-        form_next_url,
-        find_redirect(config_key),
+        get_url(form_next) if form_next else None,
+        get_url(arg_next) if arg_next else None,
+        fallback_url,
     ]
-    if declared:
-        urls.insert(0, declared)
     for url in urls:
         if url and validate_redirect_url(url):
             return url
     raise ValueError("No valid redirect URL found - configuration error")
 
 
-def get_post_login_redirect(declared: t.Optional[str] = None) -> str:
-    return get_post_action_redirect("SECURITY_POST_LOGIN_VIEW", declared)
-
-
-def get_post_register_redirect(declared: t.Optional[str] = None) -> str:
-    return get_post_action_redirect("SECURITY_POST_REGISTER_VIEW", declared)
-
-
-def get_post_logout_redirect(declared: t.Optional[str] = None) -> str:
-    return get_post_action_redirect("SECURITY_POST_LOGOUT_VIEW", declared)
-
-
-def get_post_verify_redirect(declared: t.Optional[str] = None) -> str:
-    return get_post_action_redirect("SECURITY_POST_VERIFY_VIEW", declared)
-
-
-def find_redirect(key: str) -> t.Optional[str]:
-    """Returns the URL to redirect to after a user logs in successfully.
-
-    :param key: The session or application configuration key to search for
+def simplify_url(base_url: str, redirect_url: str) -> str:
     """
-    session_value = session.pop(key.lower(), None)
-    session_url = None
-    if session_value:
-        session_url = get_url(session_value)
-    app_value = current_app.config[key.upper()]
-    app_url = None
-    if app_value:
-        app_url = get_url(app_value)
-    rv = session_url or app_url or current_app.config.get("APPLICATION_ROOT", "/")
-    return rv
+    Reduces the scheme and host from the redirect_url so it can be passed
+    as a relative URL in a query (e.g. next) param.
+    For this method we aren't worrying about a valid url (e.g. if it points
+    externally) - that will be handled by later requests.
 
-
-def propagate_next(url: str) -> str:
-    # return either URL or, if URL already has a ?next=xx, return that.
-    url_next = urlsplit(url)
-    qparams = parse_qs(url_next.query)
-    if "next" in qparams:
-        return qparams["next"][0]
-    return url
-
-
-def get_config(app: "Flask") -> t.Dict[str, t.Any]:
-    """Conveniently get the security configuration for the specified
-    application without the annoying 'SECURITY_' prefix.
-
-    :param app: The application to inspect
+    :param base_url: The URL to simplify 'against'.
+    :param redirect_url: The URL to reduce.
     """
-    items = app.config.items()
-    prefix = "SECURITY_"
+    b_url = urlsplit(base_url)
+    r_url = urlsplit(redirect_url)
 
-    def strip_prefix(tup):
-        return tup[0].replace("SECURITY_", ""), tup[1]
-
-    return dict([strip_prefix(i) for i in items if i[0].startswith(prefix)])
+    if (not r_url.scheme or r_url.scheme == b_url.scheme) and (
+        not r_url.netloc or r_url.netloc == b_url.netloc
+    ):
+        return urlunsplit(("", "", r_url.path, r_url.query, r_url.fragment))
+    return redirect_url
 
 
-def get_message(key: str, **kwargs: t.Any) -> t.Tuple[str, str]:
+def get_message(key: str, **kwargs: t.Any) -> tuple[str, str]:
     rv = config_value("MSG_" + key)
     return localize_callback(rv[0], **kwargs), rv[1]
 
@@ -673,7 +767,7 @@ def get_within_delta(key, app=None):
     """
     txt = config_value(key, app=app)
     values = txt.split()
-    return datetime.timedelta(**{values[1]: int(values[0])})
+    return timedelta(**{values[1]: int(values[0])})
 
 
 def send_mail(subject, recipient, template, **context):
@@ -701,7 +795,7 @@ def send_mail(subject, recipient, template, **context):
 
     subject = localize_callback(subject)
 
-    sender = _security.email_sender
+    sender = config_value("EMAIL_SENDER")
     if isinstance(sender, LocalProxy):
         sender = sender._get_current_object()
 
@@ -758,8 +852,8 @@ def get_token_status(token, serializer, max_age=None, return_data=False):
 
 
 def check_and_get_token_status(
-    token: str, serializer_name: str, within: datetime.timedelta
-) -> t.Tuple[bool, bool, t.Any]:
+    token: str, serializer_name: str, within: timedelta
+) -> tuple[bool, bool, t.Any]:
     """Get the status of a token and return data.
 
     :param token: The token to check
@@ -788,7 +882,7 @@ def check_and_get_token_status(
     return expired, invalid, data
 
 
-def get_identity_attributes(app: t.Optional["Flask"] = None) -> t.List[str]:
+def get_identity_attributes(app: Flask | None = None) -> list[str]:
     # Return list of keys of identity attributes
     # Is it possible to not have any?
     app = app or current_app
@@ -798,9 +892,7 @@ def get_identity_attributes(app: t.Optional["Flask"] = None) -> t.List[str]:
     return []
 
 
-def get_identity_attribute(
-    attr: str, app: t.Optional["Flask"] = None
-) -> t.Dict[str, t.Any]:
+def get_identity_attribute(attr: str, app: Flask | None = None) -> dict[str, t.Any]:
     """Given an user_identity_attribute, return the defining dict.
     A bit annoying since USER_IDENTITY_ATTRIBUTES is a list of dict
     where each dict has just one key.
@@ -819,9 +911,9 @@ def get_identity_attribute(
 def lookup_identity(identity):
     """
     Lookup identity in DB.
-    This loops through, in order, SECURITY_USER_IDENTITY_ATTRIBUTES, and first
-    calls the mapper function to validate/normalize. Then the db.find_user is called
-    on the specified user model attribute.
+    This loops through, in order, :py:data:`SECURITY_USER_IDENTITY_ATTRIBUTES`,
+    and first calls the mapper function to validate/normalize.
+    Then the db.find_user is called on the specified user model attribute.
     """
     for mapping in config_value("USER_IDENTITY_ATTRIBUTES"):
         attr = list(mapping.keys())[0]
@@ -835,7 +927,7 @@ def lookup_identity(identity):
     return None
 
 
-def uia_phone_mapper(identity: str) -> t.Optional[str]:
+def uia_phone_mapper(identity: str) -> str | None:
     """Used to match identity as a phone number. This is a simple proxy
     to :py:class:`PhoneUtil`
 
@@ -847,7 +939,7 @@ def uia_phone_mapper(identity: str) -> t.Optional[str]:
     return ph
 
 
-def uia_email_mapper(identity: str) -> t.Optional[str]:
+def uia_email_mapper(identity: str) -> str | None:
     """Used to match identity as an email.
 
     :return: Normalized email or None if not valid email.
@@ -863,7 +955,7 @@ def uia_email_mapper(identity: str) -> t.Optional[str]:
         return None
 
 
-def uia_username_mapper(identity: str) -> t.Optional[str]:
+def uia_username_mapper(identity: str) -> str | None:
     """Used to match identity as a username. This is a simple proxy
     to :py:class:`UsernameUtil`
 
@@ -877,18 +969,18 @@ def uia_username_mapper(identity: str) -> t.Optional[str]:
 def use_double_hash(password_hash=None):
     """Return a bool indicating whether a password should be hashed twice."""
     # Default to plaintext for backward compatibility with
-    # SECURITY_PASSWORD_SINGLE_HASH = False
+    # :py:data:`SECURITY_PASSWORD_SINGLE_HASH` = False
     single_hash = config_value("PASSWORD_SINGLE_HASH") or {"plaintext"}
 
     if password_hash is None:
-        scheme = _security.password_hash
+        scheme = config_value("PASSWORD_HASH")
     else:
         scheme = _pwd_context.identify(password_hash)
 
     return not (single_hash is True or scheme in single_hash)
 
 
-def csrf_cookie_handler(response: "Response") -> "Response":
+def csrf_cookie_handler(response: Response) -> Response:
     """Called at end of every request.
     Uses session to track state (set/clear)
 
@@ -919,17 +1011,14 @@ def csrf_cookie_handler(response: "Response") -> "Response":
             and session.get("remember") != "clear"
         )
         # Set cookie if successfully logged in with flask_login's remember cookie
-        if has_remember_cookie and current_user.is_authenticated:
+        if has_remember_cookie and is_user_authenticated(current_user):
             op = "set"
         else:
             return response
 
     if op == "clear":
         # Alas delete_cookie only accepts some of the keywords set_cookie does
-        # and Werkzeug didn't accept samesite, secure, httponly until 2.0
         allowed = ["path", "domain", "secure", "httponly", "samesite"]
-        if parse_version(werkzeug_version) < parse_version("2.0.0"):  # pragma: no cover
-            allowed = ["path", "domain"]
         args = {k: csrf_cookie.get(k) for k in allowed if k in csrf_cookie}
         response.delete_cookie(csrf_cookie_name, **args)
         session.pop("fs_cc")
@@ -960,18 +1049,18 @@ def csrf_cookie_handler(response: "Response") -> "Response":
 
 
 def base_render_json(
-    form: "FlaskForm",
+    form: FlaskForm,
     include_user: bool = True,
     include_auth_token: bool = False,
-    additional: t.Optional[t.Dict[str, t.Any]] = None,
+    additional: dict[str, t.Any] | None = None,
     error_status_code: int = 400,
-) -> "ResponseValue":
+) -> ResponseValue:
     """
     This method is called by all views that return JSON responses.
     This fills in the response and then calls :meth:`.Security.render_json`
     which can be overridden by the app.
     """
-    user = form.user if hasattr(form, "user") else None
+    user = getattr(form, "user", None)
     if form.errors:
         code = error_status_code
         # wtforms 3.0 introduces form-level errors - these show up as part of the
@@ -1012,8 +1101,8 @@ def base_render_json(
 
 
 def simple_render_json(
-    additional: t.Optional[t.Dict[str, t.Any]] = None,
-) -> "ResponseValue":
+    additional: dict[str, t.Any] | None = None,
+) -> ResponseValue:
     payload = dict(csrf_token=csrf.generate_csrf())
     if additional:
         payload.update(additional)
@@ -1031,7 +1120,7 @@ def default_want_json(req):
     # TODO should this handle json sub-types?
     accept_mimetypes = req.accept_mimetypes
     if not hasattr(req.accept_mimetypes, "best"):  # pragma: no cover
-        # Alright. we dont have the best property, lets add it ourselves.
+        # Alright. we don't have the best property, lets add it ourselves.
         # This is for quart compatibility
         accept_mimetypes.best = best
     if accept_mimetypes.best == "application/json":
@@ -1040,9 +1129,9 @@ def default_want_json(req):
 
 
 def json_error_response(
-    errors: t.Optional[t.Union[str, list]] = None,
-    field_errors: t.Optional[t.Dict[str, list]] = None,
-) -> t.Dict[str, t.Any]:
+    errors: str | list | None = None,
+    field_errors: dict[str | None, list] | None = None,
+) -> dict[str, t.Any]:
     """Helper to create an error response.
 
     The "errors" key holds a simple list of errors - which is made up of any passed
@@ -1051,9 +1140,11 @@ def json_error_response(
 
     The "field_errors" key which is exactly what is returned from WTForms - namely
     a dict of field-name: msg. For form-level errors (WTForms 3.0) the 'field-name' is
-    'None'
+    None - which alas means it isn't sortable and Flask's default JSONProvider
+    sorts keys - so we change that to '__all__' which is what django uses
+    apparently and was suggested as part of WTForms 3.0.
     """
-    response_json: t.Dict[str, t.Union[list, t.Dict[str, list]]] = dict()
+    response_json: dict[str, list | dict[str, list]] = dict()
     plain_errors = []
     if errors:
         if isinstance(errors, str):
@@ -1067,7 +1158,14 @@ def json_error_response(
         # we return that, as well as create a simple list of errors.
         for e in field_errors.values():
             plain_errors.extend(e)
-        response_json["field_errors"] = field_errors
+        if None in field_errors.keys():
+            # Ugh - wtforms decided to use None as a key - which json
+            # a) can't sort
+            # b) converts to "null"
+            # Issue filed - maybe they will change it
+            field_errors[""] = field_errors[None]
+            del field_errors[None]
+        response_json["field_errors"] = field_errors  # type: ignore
     response_json["errors"] = plain_errors
 
     return response_json
@@ -1096,7 +1194,7 @@ class DummySmsSender(SmsSenderBaseClass):
 
 
 class SmsSenderFactory:
-    senders: t.Dict[str, t.Type[SmsSenderBaseClass]] = {"Dummy": DummySmsSender}
+    senders: dict[str, t.Type[SmsSenderBaseClass]] = {"Dummy": DummySmsSender}
 
     @classmethod
     def createSender(cls, name, *args, **kwargs):
@@ -1128,7 +1226,7 @@ except Exception:
     pass
 
 
-def password_length_validator(password: str) -> t.Optional[t.List[str]]:
+def password_length_validator(password: str) -> list[str] | None:
     """Test password for length.
 
     :param password: Plain text password to check
@@ -1150,7 +1248,7 @@ def password_length_validator(password: str) -> t.Optional[t.List[str]]:
 
 def password_complexity_validator(
     password: str, is_register: bool, **kwargs: t.Any
-) -> t.Optional[t.List[str]]:
+) -> list[str] | None:
     """Test password for complexity.
 
     Currently just supports 'zxcvbn'.
@@ -1171,7 +1269,7 @@ def password_complexity_validator(
     if config_value("PASSWORD_COMPLEXITY_CHECKER") == "zxcvbn":
         import zxcvbn
 
-        user_info: t.List[t.Any] = []
+        user_info: list[t.Any] = []
         if not is_register:
             for v in kwargs["user"].__dict__.values():
                 if v and isinstance(v, str):
@@ -1193,7 +1291,7 @@ def password_complexity_validator(
         return None
 
 
-def password_breached_validator(password: str) -> t.Optional[t.List[str]]:
+def password_breached_validator(password: str) -> list[str] | None:
     """Check if password on breached list.
     Does nothing unless :py:data:`SECURITY_PASSWORD_CHECK_BREACHED` is set.
     If password is found on the breached list, return an error if the count is
@@ -1206,8 +1304,7 @@ def password_breached_validator(password: str) -> t.Optional[t.List[str]]:
 
     .. versionadded:: 3.4.0
     """
-    pwn = config_value("PASSWORD_CHECK_BREACHED")
-    if pwn:
+    if pwn := config_value("PASSWORD_CHECK_BREACHED"):
         try:
             cnt = pwned(password)
             if cnt >= config_value("PASSWORD_BREACHED_COUNT"):
