@@ -5,7 +5,7 @@
     Flask-Security views module
 
     :copyright: (c) 2012 by Matt Wright.
-    :copyright: (c) 2019-2023 by J. Christopher Wagner (jwag).
+    :copyright: (c) 2019-2024 by J. Christopher Wagner (jwag).
     :license: MIT, see LICENSE for more details.
 
     CSRF is tricky. By default all our forms have CSRF protection built in via
@@ -49,10 +49,16 @@ from .confirmable import (
 )
 from .decorators import anonymous_user_required, auth_required, unauth_csrf
 from .forms import (
+    _setup_methods_xlate,
+    ChangePasswordForm,
     DummyForm,
+    ForgotPasswordForm,
+    LoginForm,
     build_form_from_request,
     build_form,
     form_errors_munge,
+    ResetPasswordForm,
+    SendConfirmationForm,
     TwoFactorVerifyCodeForm,
     TwoFactorSetupForm,
     TwoFactorRescueForm,
@@ -101,7 +107,9 @@ from .utils import (
     get_request_attr,
     get_url,
     hash_password,
+    is_user_authenticated,
     json_error_response,
+    localize_callback,
     login_user,
     logout_user,
     propagate_next,
@@ -153,8 +161,9 @@ def login() -> "ResponseValue":
     access user info and csrf-token.
     For POST - redirects to POST_LOGIN_VIEW (forms) or returns 400 (json).
     """
+    form = t.cast(LoginForm, build_form_from_request("login_form"))
 
-    if current_user.is_authenticated and request.method == "POST":
+    if is_user_authenticated(current_user):
         # Just redirect current_user to POST_LOGIN_VIEW.
         # While its tempting to try to logout the current user and login the
         # new requested user - that simply doesn't work with CSRF.
@@ -163,20 +172,29 @@ def login() -> "ResponseValue":
         # 'next' - which can cause infinite redirect loops
         # (see test_common::test_authenticated_loop)
         if _security._want_json(request):
-            payload = json_error_response(
-                errors=get_message("ANONYMOUS_USER_REQUIRED")[0]
-            )
-            return _security._render_json(payload, 400, None, None)
+            if request.method == "POST":
+                payload = json_error_response(
+                    errors=get_message("ANONYMOUS_USER_REQUIRED")[0]
+                )
+                return _security._render_json(payload, 400, None, None)
+            else:
+                form.user = current_user
+                return base_render_json(form)
         else:
             return redirect(get_url(cv("POST_LOGIN_VIEW")))
 
-    form = build_form_from_request("login_form")
+    # Clean out any potential old session info - in case of previous
+    # aborted 2FA attempt.
+    tf_clean_session()
 
     if form.validate_on_submit():
         assert form.user is not None
         remember_me = form.remember.data if "remember" in form else None
         response = _security.two_factor_plugins.tf_enter(
-            form.user, remember_me, "password", next_loc=propagate_next(request.url)
+            form.user,
+            remember_me,
+            "password",
+            next_loc=propagate_next(request.url, form),
         )
         if response:
             return response
@@ -188,12 +206,8 @@ def login() -> "ResponseValue":
             return base_render_json(form, include_auth_token=True)
         return redirect(get_post_login_redirect())
 
-    if (
-        request.method == "POST"
-        and cv("RETURN_GENERIC_RESPONSES")
-        and not form.user_authenticated
-    ):
-        # Validation failed - make sure all error messages are generic
+    if request.method == "POST" and cv("RETURN_GENERIC_RESPONSES"):
+        # Validation failed - make sure PII error messages are generic
         fields_to_squash = dict(
             email=dict(replace_msg="GENERIC_AUTHN_FAILED"),
             password=dict(replace_msg="GENERIC_AUTHN_FAILED"),
@@ -202,19 +216,18 @@ def login() -> "ResponseValue":
             fields_to_squash["username"] = dict(replace_msg="GENERIC_AUTHN_FAILED")
         form_errors_munge(form, fields_to_squash)
 
-    if current_user.is_authenticated:
-        form.user = current_user
-        if _security._want_json(request):
-            return base_render_json(form)
-        return redirect(get_url(cv("POST_LOGIN_VIEW")))
-
     if _security._want_json(request):
         payload = {
             "identity_attributes": get_identity_attributes(),
         }
         return base_render_json(form, additional=payload)
 
-    if form.requires_confirmation and cv("REQUIRES_CONFIRMATION_ERROR_VIEW"):
+    if (
+        form.requires_confirmation
+        and cv("REQUIRES_CONFIRMATION_ERROR_VIEW")
+        and not cv("RETURN_GENERIC_RESPONSES")
+    ):
+        assert form.user_authenticated
         do_flash(*get_message("CONFIRMATION_REQUIRED"))
         return redirect(
             get_url(
@@ -267,7 +280,7 @@ def logout():
     """View function which handles a logout request."""
     tf_clean_session()
 
-    if current_user.is_authenticated:
+    if is_user_authenticated(current_user):
         logout_user()
 
     # No body is required - so if a POST and json - return OK
@@ -278,6 +291,7 @@ def logout():
 
 
 @anonymous_user_required
+@unauth_csrf(fall_through=True)
 def register() -> "ResponseValue":
     """View function which handles a registration request."""
 
@@ -303,7 +317,7 @@ def register() -> "ResponseValue":
         # signin - we adhere to historic behavior.
         if not _security.confirmable or cv("LOGIN_WITHOUT_CONFIRMATION"):
             response = _security.two_factor_plugins.tf_enter(
-                form.user, False, "register", next_loc=propagate_next(request.url)
+                form.user, False, "register", next_loc=propagate_next(request.url, form)
             )
             if response:
                 return response
@@ -365,19 +379,17 @@ def token_login(token):
 
     if not user or invalid:
         m, c = get_message("INVALID_LOGIN_TOKEN")
-        if _security.redirect_behavior == "spa":
-            return redirect(get_url(_security.login_error_view, qparams={c: m}))
+        if cv("REDIRECT_BEHAVIOR") == "spa":
+            return redirect(get_url(cv("LOGIN_ERROR_VIEW"), qparams={c: m}))
         do_flash(m, c)
         return redirect(url_for_security("login"))
     if expired:
         send_login_instructions(user)
-        m, c = get_message(
-            "LOGIN_EXPIRED", email=user.email, within=_security.login_within
-        )
-        if _security.redirect_behavior == "spa":
+        m, c = get_message("LOGIN_EXPIRED", email=user.email, within=cv("LOGIN_WITHIN"))
+        if cv("REDIRECT_BEHAVIOR") == "spa":
             return redirect(
                 get_url(
-                    _security.login_error_view,
+                    cv("LOGIN_ERROR_VIEW"),
                     qparams=user.get_redirect_qparams({c: m}),
                 )
             )
@@ -386,9 +398,9 @@ def token_login(token):
 
     login_user(user, authn_via=["token"])
     after_this_request(view_commit)
-    if _security.redirect_behavior == "spa":
+    if cv("REDIRECT_BEHAVIOR") == "spa":
         return redirect(
-            get_url(_security.post_login_view, qparams=user.get_redirect_qparams())
+            get_url(cv("POST_LOGIN_VIEW"), qparams=user.get_redirect_qparams())
         )
 
     do_flash(*get_message("PASSWORDLESS_LOGIN_SUCCESSFUL"))
@@ -398,8 +410,10 @@ def token_login(token):
 
 @unauth_csrf(fall_through=True)
 def send_confirmation():
-    """View function which sends confirmation instructions."""
-    form = build_form_from_request("send_confirmation_form")
+    """View function which sends confirmation instructions (/confirm)."""
+    form = t.cast(
+        SendConfirmationForm, build_form_from_request("send_confirmation_form")
+    )
 
     if form.validate_on_submit():
         send_confirmation_instructions(form.user)
@@ -410,10 +424,12 @@ def send_confirmation():
         # Here on GET or failed validate
         rinfo = dict(email=dict())
         form_errors_munge(form, rinfo)  # by suppressing errors JSON should return 200
-
-        # Make look exactly like successful (e.g. real user) request
-        if not _security._want_json(request):
-            do_flash(*get_message("CONFIRMATION_REQUEST", email=form.email.data))
+        # Check for other errors - for default form - there aren't additional fields
+        # but applications might add some (e.g. recaptcha)
+        if not form.errors:
+            # Make look exactly like successful (e.g. real user) request
+            if not _security._want_json(request):
+                do_flash(*get_message("CONFIRMATION_REQUEST", email=form.email.data))
 
     if _security._want_json(request):
         # Never include user info since this is an anonymous endpoint.
@@ -427,76 +443,80 @@ def send_confirmation():
 
 
 def confirm_email(token):
-    """View function which handles a email confirmation request."""
+    """
+    View function which handles an email confirmation request.
+    This is always a GET from an email - so for 'spa' must always redirect.
+    """
 
     expired, invalid, user = confirm_email_token_status(token)
 
-    if not user or invalid:
-        m, c = get_message("INVALID_CONFIRMATION_TOKEN")
-        if _security.redirect_behavior == "spa":
-            return redirect(get_url(_security.confirm_error_view, qparams={c: m}))
+    if not user or invalid or expired:
+        if expired:
+            m, c = get_message(
+                "CONFIRMATION_EXPIRED",
+                within=cv("CONFIRM_EMAIL_WITHIN"),
+            )
+        else:
+            m, c = get_message("INVALID_CONFIRMATION_TOKEN")
+        if cv("REDIRECT_BEHAVIOR") == "spa":
+            return redirect(get_url(cv("CONFIRM_ERROR_VIEW"), qparams={c: m}))
         do_flash(m, c)
         return redirect(
-            get_url(_security.confirm_error_view)
-            or url_for_security("send_confirmation")
+            get_url(cv("CONFIRM_ERROR_VIEW")) or url_for_security("send_confirmation")
         )
 
     already_confirmed = user.confirmed_at is not None
+    if already_confirmed:
+        m, c = get_message("ALREADY_CONFIRMED")
 
-    if expired or already_confirmed:
-        if already_confirmed:
-            m, c = get_message("ALREADY_CONFIRMED")
-        else:
-            send_confirmation_instructions(user)
-            m, c = get_message(
-                "CONFIRMATION_EXPIRED",
-                email=user.email,
-                within=_security.confirm_email_within,
-            )
-
-        if _security.redirect_behavior == "spa":
+        if cv("REDIRECT_BEHAVIOR") == "spa":
+            # No reason to expose identity info to anyone who has the link
             return redirect(
                 get_url(
-                    _security.confirm_error_view,
-                    qparams=user.get_redirect_qparams({c: m}),
+                    cv("CONFIRM_ERROR_VIEW"),
+                    qparams={c: m},
                 )
             )
 
         do_flash(m, c)
         return redirect(
-            get_url(_security.confirm_error_view)
-            or url_for_security("send_confirmation")
+            get_url(cv("CONFIRM_ERROR_VIEW")) or url_for_security("send_confirmation")
         )
 
     confirm_user(user)
     after_this_request(view_commit)
+    m, c = get_message("EMAIL_CONFIRMED")
 
+    # ? The only case where user is logged in already would be if
+    # LOGIN_WITHOUT_CONFIRMATION
     if user != current_user:
         logout_user()
         if cv("AUTO_LOGIN_AFTER_CONFIRM"):
             # N.B. this is a (small) security risk if email went to wrong place.
-            # and you have the LOGIN_WITH_CONFIRMATION flag since in that case
+            # and you have the LOGIN_WITHOUT_CONFIRMATION flag since in that case
             # you can be logged in and doing stuff - but another person could
             # get the email.
+            # Note also this goes against OWASP recommendations.
             response = _security.two_factor_plugins.tf_enter(
-                user, False, "confirm", next_loc=propagate_next(request.url)
+                user, False, "confirm", next_loc=propagate_next(request.url, None)
             )
             if response:
+                do_flash(m, c)
                 return response
             login_user(user, authn_via=["confirm"])
 
-    m, c = get_message("EMAIL_CONFIRMED")
-    if _security.redirect_behavior == "spa":
+    if cv("REDIRECT_BEHAVIOR") == "spa":
         return redirect(
             get_url(
-                _security.post_confirm_view, qparams=user.get_redirect_qparams({c: m})
+                cv("POST_CONFIRM_VIEW"),
+                qparams=user.get_redirect_qparams({c: m}),
             )
         )
     do_flash(m, c)
     return redirect(
-        get_url(_security.post_confirm_view)
+        get_url(cv("POST_CONFIRM_VIEW"))
         or get_url(
-            _security.post_login_view if cv("AUTO_LOGIN_AFTER_CONFIRM") else ".login"
+            cv("POST_LOGIN_VIEW") if cv("AUTO_LOGIN_AFTER_CONFIRM") else ".login"
         )
     )
 
@@ -504,8 +524,8 @@ def confirm_email(token):
 @anonymous_user_required
 @unauth_csrf(fall_through=True)
 def forgot_password():
-    """View function that handles a forgotten password request."""
-    form = build_form_from_request("forgot_password_form")
+    """View function that handles a forgotten password request (/reset)."""
+    form = t.cast(ForgotPasswordForm, build_form_from_request("forgot_password_form"))
 
     if form.validate_on_submit():
         send_reset_password_instructions(form.user)
@@ -513,14 +533,17 @@ def forgot_password():
             do_flash(*get_message("PASSWORD_RESET_REQUEST", email=form.email.data))
 
     elif request.method == "POST" and cv("RETURN_GENERIC_RESPONSES"):
-        # Here on GET or failed validate
+        # Here on failed validate (POST) and want generic responses
         rinfo = dict(email=dict())
         form_errors_munge(form, rinfo)  # by suppressing errors JSON should return 200
-
-        # Make look exactly like successful (e.g. real user) request
-        hash_password("not-a-password")  # reduce timing between successful and not.
-        if not _security._want_json(request):
-            do_flash(*get_message("PASSWORD_RESET_REQUEST", email=form.email.data))
+        # Check for other errors - for default form - there aren't additional fields
+        # but applications might add some (e.g. recaptcha)
+        if not form.errors:
+            # No OTHER errors on form.
+            # Make look exactly like successful (e.g. real user) request
+            hash_password("not-a-password")  # reduce timing between successful and not.
+            if not _security._want_json(request):
+                do_flash(*get_message("PASSWORD_RESET_REQUEST", email=form.email.data))
 
     if _security._want_json(request):
         # Never include user info since this is an anonymous endpoint.
@@ -528,13 +551,13 @@ def forgot_password():
 
     if (
         form.requires_confirmation
-        and _security.requires_confirmation_error_view
+        and cv("REQUIRES_CONFIRMATION_ERROR_VIEW")
         and not cv("RETURN_GENERIC_RESPONSES")
     ):
         do_flash(*get_message("CONFIRMATION_REQUIRED"))
         return redirect(
             get_url(
-                _security.requires_confirmation_error_view,
+                cv("REQUIRES_CONFIRMATION_ERROR_VIEW"),
                 qparams={"email": form.email.data},
             )
         )
@@ -549,7 +572,7 @@ def forgot_password():
 @anonymous_user_required
 @unauth_csrf(fall_through=True)
 def reset_password(token):
-    """View function that handles a reset password request.
+    """View function that handles a reset password request (/reset/<token>).
 
     This is usually called via GET as part of an email link and redirects to
     a reset-password form
@@ -559,46 +582,39 @@ def reset_password(token):
     the 'forgot-password' form.
 
     In the case of non-form based configuration:
-    For GET normal case - redirect to RESET_VIEW?token={token}&email={email}
-    For GET invalid case - redirect to RESET_ERROR_VIEW?error={error}&email={email}
+    For GET normal case - redirect to RESET_VIEW?token={token}
+    For GET invalid case - redirect to RESET_ERROR_VIEW?error={error}
     For POST normal/successful case - return 200 with new authentication token
-    For POST error case return 400 with form.errors
+    For POST error case return 400
     """
 
     expired, invalid, user = reset_password_token_status(token)
-    form = build_form_from_request("reset_password_form")
+    form = t.cast(ResetPasswordForm, build_form_from_request("reset_password_form"))
     form.user = user
 
     if request.method == "GET":
-        if not user or invalid:
-            m, c = get_message("INVALID_RESET_PASSWORD_TOKEN")
-            if _security.redirect_behavior == "spa":
-                return redirect(get_url(_security.reset_error_view, qparams={c: m}))
-            do_flash(m, c)
-            return redirect(url_for_security("forgot_password"))
-        if expired:
-            send_reset_password_instructions(user)
-            m, c = get_message(
-                "PASSWORD_RESET_EXPIRED",
-                email=user.email,
-                within=_security.reset_password_within,
-            )
-            if _security.redirect_behavior == "spa":
-                return redirect(
-                    get_url(
-                        _security.reset_error_view,
-                        qparams=user.get_redirect_qparams({c: m}),
-                    )
+        if not user or invalid or expired:
+            if expired:
+                m, c = get_message(
+                    "PASSWORD_RESET_EXPIRED",
+                    within=cv("RESET_PASSWORD_WITHIN"),
                 )
+            else:
+                m, c = get_message("INVALID_RESET_PASSWORD_TOKEN")
+            if cv("REDIRECT_BEHAVIOR") == "spa":
+                return redirect(get_url(cv("RESET_ERROR_VIEW"), qparams={c: m}))
             do_flash(m, c)
             return redirect(url_for_security("forgot_password"))
 
         # All good - for SPA - redirect to the ``reset_view``
-        if _security.redirect_behavior == "spa":
+        # Still - don't include PII such as identity and email if someone
+        # intercepts link they still won't necessarily know the login identity
+        # (even though they can change the password!).
+        if cv("REDIRECT_BEHAVIOR") == "spa":
             return redirect(
                 get_url(
-                    _security.reset_view,
-                    qparams=user.get_redirect_qparams({"token": token}),
+                    cv("RESET_VIEW"),
+                    qparams={"token": token},
                 )
             )
         # for forms - render the reset password form
@@ -610,49 +626,48 @@ def reset_password(token):
         )
 
     # This is the POST case.
-    m = None
-    if not user or invalid:
-        invalid = True
-        m, c = get_message("INVALID_RESET_PASSWORD_TOKEN")
-        if not _security._want_json(request):
-            do_flash(m, c)
+    if not user or invalid or expired:
+        if expired:
+            m, c = get_message(
+                "PASSWORD_RESET_EXPIRED", within=cv("RESET_PASSWORD_WITHIN")
+            )
+        else:
+            m, c = get_message("INVALID_RESET_PASSWORD_TOKEN")
 
-    if expired:
-        send_reset_password_instructions(user)
-        m, c = get_message(
-            "PASSWORD_RESET_EXPIRED",
-            email=user.email,
-            within=_security.reset_password_within,
-        )
-        if not _security._want_json(request):
-            do_flash(m, c)
-
-    if invalid or expired:
         if _security._want_json(request):
             form.form_errors.append(m)
             return base_render_json(form, include_user=False)
         else:
+            do_flash(m, c)
             return redirect(url_for_security("forgot_password"))
 
     if form.validate_on_submit():
         after_this_request(view_commit)
         update_password(user, form.password.data)
-        response = _security.two_factor_plugins.tf_enter(
-            form.user, False, "reset", next_loc=propagate_next(request.url)
-        )
-        if response:
-            return response
-        # two factor not required - just login
-        login_user(user, authn_via=["reset"])
-        if _security._want_json(request):
-            dummy_form = DummyForm(formdata=None)
-            dummy_form.user = user
-            return base_render_json(dummy_form, include_auth_token=True)
-        else:
-            do_flash(*get_message("PASSWORD_RESET"))
-            return redirect(
-                get_url(_security.post_reset_view) or get_url(_security.post_login_view)
+        if cv("AUTO_LOGIN_AFTER_RESET"):
+            # backwards compat - really shouldn't do this according to OWASP
+            response = _security.two_factor_plugins.tf_enter(
+                form.user, False, "reset", next_loc=propagate_next(request.url, None)
             )
+            if response:
+                return response
+            # two factor not required - just login
+            login_user(user, authn_via=["reset"])
+            if _security._want_json(request):
+                dummy_form = DummyForm(formdata=None)
+                dummy_form.user = user
+                return base_render_json(dummy_form, include_auth_token=True)
+            else:
+                do_flash(*get_message("PASSWORD_RESET"))
+                return redirect(
+                    get_url(cv("POST_RESET_VIEW")) or get_url(cv("POST_LOGIN_VIEW"))
+                )
+        else:
+            if _security._want_json(request):
+                return _security._render_json({}, 200, None, None)
+            else:
+                do_flash(*get_message("PASSWORD_RESET_NO_LOGIN"))
+                return redirect(get_url(cv("POST_RESET_VIEW")) or get_url(".login"))
 
     # validation failure case - for forms - we try again including the token
     # for non-forms -  we just return errors and assume caller remembers token.
@@ -669,7 +684,7 @@ def reset_password(token):
 @auth_required(lambda: cv("API_ENABLED_METHODS"))
 def change_password():
     """View function which handles a change password request."""
-    form = build_form_from_request("change_password_form")
+    form = t.cast(ChangePasswordForm, build_form_from_request("change_password_form"))
 
     if not current_user.password:
         # This is case where user registered w/o a password - since we can't
@@ -693,7 +708,7 @@ def change_password():
 
         do_flash(*get_message("PASSWORD_CHANGE"))
         return redirect(
-            get_url(_security.post_change_view) or get_url(_security.post_login_view)
+            get_url(cv("POST_CHANGE_VIEW")) or get_url(cv("POST_LOGIN_VIEW"))
         )
 
     active_password = True if current_user.password else False
@@ -732,7 +747,7 @@ def two_factor_setup():
     """
     form = t.cast(TwoFactorSetupForm, build_form_from_request("two_factor_setup_form"))
 
-    if not current_user.is_authenticated:
+    if not is_user_authenticated(current_user):
         # This is the initial login case
         # We can also get here from setup if they want to change TODO: how?
         if not all(k in session for k in ["tf_user_id", "tf_state"]) or session[
@@ -749,6 +764,7 @@ def two_factor_setup():
 
     else:
         # Caller is changing their TFA profile. This requires a 'fresh' authentication
+        # N.B unauth_csrf has done the CSRF check already.
         if not check_and_update_authn_fresh(
             cv("FRESHNESS"),
             cv("FRESHNESS_GRACE_PERIOD"),
@@ -773,38 +789,34 @@ def two_factor_setup():
             after_this_request(view_commit)
             if not _security._want_json(request):
                 do_flash(*get_message("TWO_FACTOR_DISABLED"))
-                return redirect(get_url(_security.post_login_view))
+                return redirect(get_url(cv("TWO_FACTOR_POST_SETUP_VIEW")))
             else:
                 return base_render_json(form)
 
-        # Regenerate the TOTP secret on every call of 2FA setup unless it is
-        # within the same session and method (e.g. upon entering the phone number)
-        if pm != session.get("tf_primary_method", None):
-            session["tf_totp_secret"] = _security._totp_factory.generate_totp_secret()
-
+        # Regenerate the TOTP secret on every call of 2FA setup
+        totp = _security._totp_factory.generate_totp_secret()
+        phone = form.phone.data if pm == "sms" else None
+        session["tf_totp_secret"] = totp
         session["tf_primary_method"] = pm
         session["tf_state"] = "validating_profile"
         json_response = {
             "tf_state": "validating_profile",
-            "tf_primary_method": pm,
+            "tf_primary_method": pm,  # old
+            "tf_method": pm,
         }
-        new_phone = (
-            form.phone.data if (form.phone.data and len(form.phone.data) > 0) else None
-        )
-        if new_phone:
-            user.tf_phone_number = new_phone
+        if phone:
+            #  TODO dont save here - wait until complete
+            user.tf_phone_number = phone
             _datastore.put(user)
             after_this_request(view_commit)
 
-        # This form is sort of bizarre - for SMS and authenticator
-        # you select, then get more info, and submit again.
-        # For authenticator of course, we don't actually send anything
-        # and for SMS it is the second time around that we get the phone number
-        if pm == "email" or (pm == "sms" and new_phone):
+        if (
+            pm == "email" or pm == "sms"
+        ):  # TODO not sure this is needed - send checks this
             msg = user.tf_send_security_token(
                 method=pm,
-                totp_secret=session["tf_totp_secret"],
-                phone_number=getattr(user, "tf_phone_number", None),
+                totp_secret=totp,
+                phone_number=phone,
             )
             if msg:
                 # send code didn't work
@@ -817,9 +829,7 @@ def two_factor_setup():
 
         qrcode_values = dict()
         if pm == "authenticator":
-            authr_setup_values = _security._totp_factory.fetch_setup_values(
-                session["tf_totp_secret"], user
-            )
+            authr_setup_values = _security._totp_factory.fetch_setup_values(totp, user)
             # Add all the values used in qrcode to json response
             json_response["tf_authr_key"] = authr_setup_values["key"]
             json_response["tf_authr_username"] = authr_setup_values["username"]
@@ -831,31 +841,35 @@ def two_factor_setup():
                 authr_username=authr_setup_values["username"],
                 authr_issuer=authr_setup_values["issuer"],
             )
+        if _security._want_json(request):
+            return base_render_json(form, include_user=False, additional=json_response)
         code_form = build_form("two_factor_verify_code_form")
-        if not _security._want_json(request):
-            return _security.render_template(
-                cv("TWO_FACTOR_SETUP_TEMPLATE"),
-                two_factor_setup_form=form,
-                two_factor_verify_code_form=code_form,
-                choices=cv("TWO_FACTOR_ENABLED_METHODS"),
-                chosen_method=pm,
-                **qrcode_values,
-                **_ctx("tf_setup"),
-            )
-        return base_render_json(form, include_user=False, additional=json_response)
+        return _security.render_template(
+            cv("TWO_FACTOR_SETUP_TEMPLATE"),
+            two_factor_setup_form=form,
+            two_factor_verify_code_form=code_form,
+            choices=cv("TWO_FACTOR_ENABLED_METHODS"),
+            chosen_method=pm,  # do not translate
+            primary_method=localize_callback(
+                _setup_methods_xlate[getattr(user, "tf_primary_method", None)]
+            ),
+            **qrcode_values,
+            **_ctx("tf_setup"),
+        )
 
     # We get here on GET and POST with failed validation.
     # For things like phone number - we've already done one POST
     # that succeeded and now it failed - so retain the initial info
     choices = cv("TWO_FACTOR_ENABLED_METHODS")
-    if not cv("TWO_FACTOR_REQUIRED"):
-        choices.append("disable")
+    if (not cv("TWO_FACTOR_REQUIRED")) and user.tf_primary_method is not None:
+        choices.insert(0, "disable")
 
     if _security._want_json(request):
         # Provide information application/UI might need to render their own form/input
         json_response = {
             "tf_required": cv("TWO_FACTOR_REQUIRED"),
-            "tf_primary_method": getattr(user, "tf_primary_method", None),
+            "tf_primary_method": getattr(user, "tf_primary_method", None),  # old
+            "tf_method": getattr(user, "tf_primary_method", None),
             "tf_phone_number": getattr(user, "tf_phone_number", None),
             "tf_available_methods": choices,
         }
@@ -868,7 +882,9 @@ def two_factor_setup():
         two_factor_verify_code_form=code_form,
         choices=choices,
         chosen_method=form.setup.data,
-        primary_method=getattr(user, "tf_primary_method", "None"),
+        primary_method=localize_callback(
+            _setup_methods_xlate[getattr(user, "tf_primary_method", None)]
+        ),
         two_factor_required=cv("TWO_FACTOR_REQUIRED"),
         **_ctx("tf_setup"),
     )
@@ -883,70 +899,63 @@ def two_factor_token_validation():
        In this case - user not logged in -
        but 'tf_state' == 'ready' or 'validating_profile'
     2) validating after CHANGE/ENABLE 2FA. In this case user logged in/authenticated
+       In this case we allow a GET to get the specific enter-code form.
 
     """
     form = t.cast(
         TwoFactorVerifyCodeForm, build_form_from_request("two_factor_verify_code_form")
     )
 
-    changing = current_user.is_authenticated
+    # state info in session
+    pm = session.get("tf_primary_method", None)
+    totp_secret = session.get("tf_totp_secret", None)
+    tf_state = session.get("tf_state", None)
+    tf_user_id = session.get("tf_user_id", None)
+
+    changing = is_user_authenticated(current_user)
     if not changing:
-        # This is the normal login case OR initial setup
+        # This is the normal login case OR initial setup (two factor required)
         if (
-            not all(k in session for k in ["tf_user_id", "tf_state"])
-            or session["tf_state"] not in ["ready", "validating_profile"]
-            or (
-                session["tf_state"] == "validating_profile"
-                and not all(
-                    k in session for k in ["tf_primary_method", "tf_totp_secret"]
-                )
-            )
+            tf_state not in ["ready", "validating_profile"]
+            or (tf_state == "validating_profile" and not all([pm, totp_secret]))
+            or not tf_user_id
         ):
             # illegal call on this endpoint
             tf_clean_session()
             return tf_illegal_state(form, cv("TWO_FACTOR_ERROR_VIEW"))
 
-        user = _datastore.find_user(fs_uniquifier=session["tf_user_id"])
+        user = _datastore.find_user(fs_uniquifier=tf_user_id)
         form.user = user
         if not user:
             tf_clean_session()
             return tf_illegal_state(form, cv("TWO_FACTOR_ERROR_VIEW"))
 
-        if session["tf_state"] == "ready":
-            # normal login case
+        if tf_state == "ready":
+            # normal login case - use saved values
             pm = user.tf_primary_method
             totp_secret = user.tf_totp_secret
-        else:
-            # initial setup case
-            pm = session["tf_primary_method"]
-            totp_secret = session["tf_totp_secret"]
     else:
         # Changing TFA profile - user is already authenticated.
-        if (
-            not all(k in session for k in ["tf_state", "tf_primary_method"])
-            or session["tf_state"] != "validating_profile"
-        ):
+        if tf_state != "validating_profile" or not all([pm, totp_secret]):
             tf_clean_session()
             # logout since this seems like attack-ish/logic error
             logout_user()
             return tf_illegal_state(form, cv("TWO_FACTOR_ERROR_VIEW"))
-        pm = session["tf_primary_method"]
-        totp_secret = session["tf_totp_secret"]
         form.user = current_user
 
     form.primary_method = pm
     form.tf_totp_secret = totp_secret
     if form.validate_on_submit():
-        # Success - log in user and clear all session variables
+        # Success - finish process based on 'changing' and clear all session variables
         completion_message, token = complete_two_factor_process(
             form.user, pm, totp_secret, changing
         )
 
         after_this_request(view_commit)
+        if token:
+            after_this_request(partial(tf_set_validity_token_cookie, token=token))
 
         if not _security._want_json(request):
-            if token:
-                after_this_request(partial(tf_set_validity_token_cookie, token=token))
             do_flash(*get_message(completion_message))
             if changing:
                 return redirect(get_url(cv("TWO_FACTOR_POST_SETUP_VIEW")))
@@ -954,10 +963,7 @@ def two_factor_token_validation():
                 return redirect(get_post_login_redirect())
 
         else:
-            json_payload = {}
-            if token:
-                json_payload["tf_validity_token"] = token
-            return base_render_json(form, additional=json_payload)
+            return base_render_json(form, include_auth_token=True)
 
     # GET or not successful POST
 
@@ -965,15 +971,12 @@ def two_factor_token_validation():
     if changing:
         if _security._want_json(request):
             return base_render_json(form)
-        setup_form = build_form("two_factor_setup_form")
-
+        # allow app to fetch just this form (independent of /tf_setup)
         return _security.render_template(
-            cv("TWO_FACTOR_SETUP_TEMPLATE"),
-            two_factor_setup_form=setup_form,
+            cv("TWO_FACTOR_VERIFY_CODE_TEMPLATE"),
             two_factor_verify_code_form=form,
-            chosen_method=pm,
-            choices=cv("TWO_FACTOR_ENABLED_METHODS"),
-            **_ctx("tf_setup"),
+            chosen_method=localize_callback(_setup_methods_xlate[pm]),
+            **_ctx("tf_token_validation"),
         )
 
     # if we were trying to validate an existing method
@@ -988,7 +991,7 @@ def two_factor_token_validation():
             cv("TWO_FACTOR_VERIFY_CODE_TEMPLATE"),
             two_factor_rescue_form=rescue_form,
             two_factor_verify_code_form=form,
-            chosen_method=pm,
+            chosen_method=localize_callback(_setup_methods_xlate[pm]),
             problem=None,
             **_ctx("tf_token_validation"),
         )
@@ -1056,7 +1059,9 @@ def two_factor_rescue():
         cv("TWO_FACTOR_VERIFY_CODE_TEMPLATE"),
         two_factor_verify_code_form=code_form,
         two_factor_rescue_form=form,
-        chosen_method=form.user.tf_primary_method,
+        chosen_method=localize_callback(
+            _setup_methods_xlate[form.user.tf_primary_method]
+        ),
         rescue_mail=cv("TWO_FACTOR_RESCUE_MAIL"),
         problem=rproblem,
         **_ctx("tf_token_validation"),
@@ -1067,18 +1072,20 @@ def create_blueprint(app, state, import_name):
     """Creates the security extension blueprint"""
 
     bp = Blueprint(
-        state.blueprint_name,
+        cv("BLUEPRINT_NAME", app=app),
         import_name,
-        url_prefix=state.url_prefix,
-        subdomain=state.subdomain,
+        url_prefix=cv("URL_PREFIX", app=app),
+        subdomain=cv("SUBDOMAIN", app=app),
         template_folder="templates",
         static_folder=cv("STATIC_FOLDER", app),
         static_url_path=cv("STATIC_FOLDER_URL", app),
     )
 
-    if state.logout_methods is not None:
+    if cv("LOGOUT_METHODS", app=app) is not None:
         bp.route(
-            cv("LOGOUT_URL", app=app), methods=state.logout_methods, endpoint="logout"
+            cv("LOGOUT_URL", app=app),
+            methods=cv("LOGOUT_METHODS", app=app),
+            endpoint="logout",
         )(logout)
 
     login_url = cv("LOGIN_URL", app=app)
@@ -1118,7 +1125,7 @@ def create_blueprint(app, state, import_name):
         bp.route(us_setup_url, methods=["GET", "POST"], endpoint="us_setup")(us_setup)
         bp.route(
             us_setup_url + slash_url_suffix(us_setup_url, "<token>"),
-            methods=["GET", "POST"],
+            methods=["POST"],
             endpoint="us_setup_validate",
         )(us_setup_validate)
 
