@@ -4,7 +4,7 @@
 
     Flask-Security WebAuthn module
 
-    :copyright: (c) 2021-2022 by J. Christopher Wagner (jwag).
+    :copyright: (c) 2021-2024 by J. Christopher Wagner (jwag).
     :license: MIT, see LICENSE for more details.
 
     This implements support for webauthn/FIDO2 Level 2 using the py_webauthn package.
@@ -36,16 +36,18 @@
 
 """
 
-import datetime
+from __future__ import annotations
+
 import json
 import time
 import typing as t
 from functools import partial
 
 from flask import abort, after_this_request, request, session
-from flask import current_app as app
+from flask import current_app
 from flask_login import current_user
 from wtforms import BooleanField, HiddenField, RadioField, StringField, SubmitField
+from .forms import NextFormMixin
 
 try:
     import webauthn
@@ -53,16 +55,19 @@ try:
         VerifiedAuthentication,
     )
     from webauthn.registration.verify_registration_response import VerifiedRegistration
+    from webauthn.helpers import (
+        parse_registration_credential_json,
+        parse_authentication_credential_json,
+    )
     from webauthn.helpers.exceptions import (
         InvalidAuthenticationResponse,
+        InvalidJSONStructure,
         InvalidRegistrationResponse,
     )
     from webauthn.helpers.structs import (
-        AuthenticationCredential,
         AuthenticatorTransport,
         PublicKeyCredentialDescriptor,
         PublicKeyCredentialType,
-        RegistrationCredential,
         UserVerificationRequirement,
     )
     from webauthn.helpers import bytes_to_base64url
@@ -155,8 +160,8 @@ class WebAuthnRegisterResponseForm(Form):
     usage: str
     user_verification: bool
     # this is returned to caller (not part of the client form)
-    registration_verification: "VerifiedRegistration"
-    transports: t.List[str] = []
+    registration_verification: VerifiedRegistration
+    transports: list[AuthenticatorTransport] = []
     extensions: str
 
     def validate(self, **kwargs: t.Any) -> bool:
@@ -168,8 +173,13 @@ class WebAuthnRegisterResponseForm(Form):
             self.credential.errors.append(msg)
             return False
         try:
-            reg_cred = RegistrationCredential.parse_raw(self.credential.data)
-        except (ValueError, KeyError):
+            reg_cred = parse_registration_credential_json(self.credential.data)
+        except (
+            ValueError,
+            KeyError,
+            InvalidJSONStructure,
+            InvalidRegistrationResponse,
+        ):
             self.credential.errors.append(get_message("API_ERROR")[0])
             return False
         try:
@@ -193,24 +203,22 @@ class WebAuthnRegisterResponseForm(Form):
             )
             return False
 
-        # Alas py_webauthn doesn't support extensions nor transports yet
+        self.transports = (
+            reg_cred.response.transports if reg_cred.response.transports else []
+        )
+        # Alas py_webauthn doesn't support extensions
         response_full = json.loads(self.credential.data)
         # TODO - verify this is JSON (created with JSON.stringify)
         self.extensions = response_full.get("extensions", None)
-        self.transports = (
-            [tr for tr in response_full["transports"]]
-            if response_full.get("transports", None)
-            else []
-        )
         return True
 
 
-class WebAuthnSigninForm(Form):
+class WebAuthnSigninForm(Form, NextFormMixin):
     identity = StringField(get_form_field_label("identity"))
     remember = BooleanField(get_form_field_label("remember_me"))
     submit = SubmitField(label=get_form_field_xlate(_("Start")), id="wan_signin")
 
-    user: t.Optional["User"] = None
+    user: User | None = None
     # set by caller - is this a second factor authentication?
     is_secondary: bool
 
@@ -236,7 +244,7 @@ class WebAuthnSigninForm(Form):
         return True
 
 
-class WebAuthnSigninResponseForm(Form):
+class WebAuthnSigninResponseForm(Form, NextFormMixin):
     """
     This form is used both for signin (primary/first or secondary) and verify.
     """
@@ -251,9 +259,9 @@ class WebAuthnSigninResponseForm(Form):
     is_secondary: bool
     is_verify: bool
     # returned to caller
-    authentication_verification: "VerifiedAuthentication"
-    user: t.Optional["User"] = None
-    cred: t.Optional["WebAuthn"] = None
+    authentication_verification: VerifiedAuthentication
+    user: User | None = None
+    cred: WebAuthn | None = None
     # Set to True if this authentication qualifies as 'multi-factor'
     mf_check: bool = False
 
@@ -261,8 +269,13 @@ class WebAuthnSigninResponseForm(Form):
         if not super().validate(**kwargs):
             return False  # pragma: no cover
         try:
-            auth_cred = AuthenticationCredential.parse_raw(self.credential.data)
-        except (ValueError, KeyError):
+            auth_cred = parse_authentication_credential_json(self.credential.data)
+        except (
+            ValueError,
+            KeyError,
+            InvalidJSONStructure,
+            InvalidAuthenticationResponse,
+        ):
             self.credential.errors.append(get_message("API_ERROR")[0])
             return False
 
@@ -360,7 +373,7 @@ class WebAuthnDeleteForm(Form):
 class WebAuthnVerifyForm(Form):
     submit = SubmitField(label=get_form_field_label("submit"), id="wan_verify")
 
-    user: "User"
+    user: User
 
     def validate(self, **kwargs: t.Any) -> bool:
         if not super().validate(**kwargs):
@@ -375,7 +388,7 @@ class WebAuthnVerifyForm(Form):
     within=lambda: cv("FRESHNESS"),
     grace=lambda: cv("FRESHNESS_GRACE_PERIOD"),
 )
-def webauthn_register() -> "ResponseValue":
+def webauthn_register() -> ResponseValue:
     """Start Registration for an existing authenticated user
 
     Note that it requires a POST to start the registration and must send 'name'
@@ -384,9 +397,11 @@ def webauthn_register() -> "ResponseValue":
     Also - this requires that the user already be logged in - so we can provide info
     as part of the GET that could otherwise be considered leaking user info.
     """
-    payload: t.Dict[str, t.Any]
+    payload: dict[str, t.Any]
 
-    form: WebAuthnRegisterForm = build_form_from_request("wan_register_form")
+    form: WebAuthnRegisterForm = t.cast(
+        WebAuthnRegisterForm, build_form_from_request("wan_register_form")
+    )
 
     if form.validate_on_submit():
         challenge = _security._webauthn_util.generate_challenge(
@@ -403,7 +418,7 @@ def webauthn_register() -> "ResponseValue":
             challenge=challenge.encode(),
             rp_name=cv("WAN_RP_NAME"),
             rp_id=request.host.split(":")[0],
-            user_id=current_user.fs_webauthn_user_handle,
+            user_id=current_user.fs_webauthn_user_handle.encode(),
             user_name=current_user.calc_username(),
             timeout=cv("WAN_REGISTER_TIMEOUT"),
             exclude_credentials=create_credential_list(
@@ -445,11 +460,11 @@ def webauthn_register() -> "ResponseValue":
             wan_register_response_form=build_form("wan_register_response_form"),
             wan_state=state_token,
             credential_options=json.dumps(co_json),
-            **_security._run_ctx_processor("wan_register")
+            **_security._run_ctx_processor("wan_register"),
         )
 
     current_creds = []
-    cred: "WebAuthn"
+    cred: WebAuthn
     for cred in current_user.webauthn:
         cl = {
             "name": cred.name,
@@ -457,12 +472,12 @@ def webauthn_register() -> "ResponseValue":
             "transports": cred.transports,
             "lastuse": cred.lastuse_datetime.isoformat(),
             "usage": cred.usage,
-            "backup_state": cred.backup_state
-            if hasattr(cred, "backup_state")
-            else False,
-            "device_type": cred.device_type
-            if hasattr(cred, "device_type")
-            else "Unknown",
+            "backup_state": (
+                cred.backup_state if hasattr(cred, "backup_state") else False
+            ),
+            "device_type": (
+                cred.device_type if hasattr(cred, "device_type") else "Unknown"
+            ),
         }
         # TODO: i18n
         discoverable = "Unknown"
@@ -481,15 +496,16 @@ def webauthn_register() -> "ResponseValue":
         wan_register_form=form,
         wan_delete_form=build_form("wan_delete_form"),
         registered_credentials=current_creds,
-        **_security._run_ctx_processor("wan_register")
+        **_security._run_ctx_processor("wan_register"),
     )
 
 
 @auth_required(lambda: cv("API_ENABLED_METHODS"))
-def webauthn_register_response(token: str) -> "ResponseValue":
+def webauthn_register_response(token: str) -> ResponseValue:
     """Response from browser."""
-    form: WebAuthnRegisterResponseForm = build_form_from_request(
-        "wan_register_response_form"
+    form: WebAuthnRegisterResponseForm = t.cast(
+        WebAuthnRegisterResponseForm,
+        build_form_from_request("wan_register_response_form"),
     )
 
     expired, invalid, state = check_and_get_token_status(
@@ -527,12 +543,13 @@ def webauthn_register_response(token: str) -> "ResponseValue":
                 "credential_device_type",
                 "single_device",
             ),
-            transports=form.transports,
+            transports=list(form.transports),
             extensions=form.extensions,
             usage=form.usage,
         )
         wan_registered.send(
-            app._get_current_object(),  # type: ignore
+            current_app._get_current_object(),  # type: ignore
+            _async_wrapper=current_app.ensure_sync,
             user=current_user,
             name=state["name"],
         )
@@ -545,12 +562,13 @@ def webauthn_register_response(token: str) -> "ResponseValue":
 
     if _security._want_json(request):
         return base_render_json(form)
-    if len(form.errors) > 0:
-        do_flash(form.errors["credential"][0], "error")
+    if form.errors:
+        for v in form.errors.values():
+            do_flash(v[0], "error")
     return redirect(url_for_security("wan_register"))
 
 
-def _signin_common(user: t.Optional["User"], usage: t.List[str]) -> t.Tuple[t.Any, str]:
+def _signin_common(user: User | None, usage: list[str]) -> tuple[t.Any, str]:
     """
     Common code between signin and verify - once form has been verified.
     """
@@ -586,7 +604,7 @@ def _signin_common(user: t.Optional["User"], usage: t.List[str]) -> t.Tuple[t.An
 
 @anonymous_user_required
 @unauth_csrf(fall_through=True)
-def webauthn_signin() -> "ResponseValue":
+def webauthn_signin() -> ResponseValue:
     # This view can be called either as a 'first' authentication or as part of
     # 2FA.
     is_secondary = all(k in session for k in ["tf_user_id", "tf_state"]) and session[
@@ -597,7 +615,7 @@ def webauthn_signin() -> "ResponseValue":
     else:
         abort(404)
 
-    form: WebAuthnSigninForm = build_form_from_request("wan_signin_form")
+    form = t.cast(WebAuthnSigninForm, build_form_from_request("wan_signin_form"))
     form.is_secondary = is_secondary
     if form.validate_on_submit():
         o_json, state_token = _signin_common(
@@ -618,12 +636,14 @@ def webauthn_signin() -> "ResponseValue":
             cv("WAN_SIGNIN_TEMPLATE"),
             wan_signin_form=form,
             wan_signin_response_form=build_form(
-                "wan_signin_response_form", remember=form.remember.data
+                "wan_signin_response_form",
+                remember=form.remember.data,
+                next=form.next.data,
             ),
             wan_state=state_token,
             credential_options=json.dumps(o_json),
             is_secondary=is_secondary,
-            **_security._run_ctx_processor("wan_signin")
+            **_security._run_ctx_processor("wan_signin"),
         )
 
     if _security._want_json(request):
@@ -633,18 +653,18 @@ def webauthn_signin() -> "ResponseValue":
         wan_signin_form=form,
         wan_signin_response_form=build_form("wan_signin_response_form"),
         is_secondary=is_secondary,
-        **_security._run_ctx_processor("wan_signin")
+        **_security._run_ctx_processor("wan_signin"),
     )
 
 
 @unauth_csrf(fall_through=True)
-def webauthn_signin_response(token: str) -> "ResponseValue":
+def webauthn_signin_response(token: str) -> ResponseValue:
     is_secondary = all(k in session for k in ["tf_user_id", "tf_state"]) and session[
         "tf_state"
     ] in ["ready"]
 
-    form: WebAuthnSigninResponseForm = build_form_from_request(
-        "wan_signin_response_form"
+    form = t.cast(
+        WebAuthnSigninResponseForm, build_form_from_request("wan_signin_response_form")
     )
 
     expired, invalid, state = check_and_get_token_status(
@@ -671,7 +691,7 @@ def webauthn_signin_response(token: str) -> "ResponseValue":
         after_this_request(view_commit)
         assert form.cred
         assert form.user
-        form.cred.lastuse_datetime = datetime.datetime.utcnow()
+        form.cred.lastuse_datetime = _security.datetime_factory()
         form.cred.sign_count = form.authentication_verification.new_sign_count
         form.cred.backup_state = getattr(
             form.authentication_verification, "credential_backed_up", False
@@ -685,7 +705,6 @@ def webauthn_signin_response(token: str) -> "ResponseValue":
         if is_secondary:
             tf_token = _security.two_factor_plugins.tf_complete(form.user, True)
             if tf_token:
-                json_payload["tf_validity_token"] = tf_token
                 after_this_request(
                     partial(tf_set_validity_token_cookie, token=tf_token)
                 )
@@ -703,7 +722,7 @@ def webauthn_signin_response(token: str) -> "ResponseValue":
                     form.user,
                     remember_me,
                     "webauthn",
-                    next_loc=propagate_next(request.url),
+                    next_loc=propagate_next(request.url, form),
                 )
                 if response:
                     return response
@@ -726,8 +745,9 @@ def webauthn_signin_response(token: str) -> "ResponseValue":
 
     # Since the response is auto submitted - we go back to
     # signin form - for now use flash.
-    if form.credential.errors:
-        do_flash(form.credential.errors[0], "error")
+    if form.errors:
+        for v in form.errors.values():
+            do_flash(v[0], "error")
     return redirect(url_for_security("wan_signin"))
 
 
@@ -736,9 +756,9 @@ def webauthn_signin_response(token: str) -> "ResponseValue":
     within=lambda: cv("FRESHNESS"),
     grace=lambda: cv("FRESHNESS_GRACE_PERIOD"),
 )
-def webauthn_delete() -> "ResponseValue":
+def webauthn_delete() -> ResponseValue:
     """Deletes an existing registered credential."""
-    form: WebAuthnDeleteForm = build_form_from_request("wan_delete_form")
+    form = t.cast(WebAuthnDeleteForm, build_form_from_request("wan_delete_form"))
 
     if form.validate_on_submit():
         # validate made sure form.name.data exists.
@@ -746,7 +766,8 @@ def webauthn_delete() -> "ResponseValue":
         after_this_request(view_commit)
 
         wan_deleted.send(
-            app._get_current_object(),  # type: ignore
+            current_app._get_current_object(),  # type: ignore
+            _async_wrapper=current_app.ensure_sync,
             user=current_user,
             name=cred.name,
         )
@@ -764,14 +785,14 @@ def webauthn_delete() -> "ResponseValue":
 
 
 @auth_required(lambda: cv("API_ENABLED_METHODS"))
-def webauthn_verify() -> "ResponseValue":
+def webauthn_verify() -> ResponseValue:
     """
     Re-authenticate to reset freshness time.
     This is likely the result of a reauthn_handler redirect, which
     will have filled in ?next=xxx - which we want to carefully not lose as we
     go through these steps.
     """
-    form: WebAuthnVerifyForm = build_form_from_request("wan_verify_form")
+    form = t.cast(WebAuthnVerifyForm, build_form_from_request("wan_verify_form"))
 
     if form.validate_on_submit():
         o_json, state_token = _signin_common(form.user, cv("WAN_ALLOW_AS_VERIFY"))
@@ -785,7 +806,7 @@ def webauthn_verify() -> "ResponseValue":
             wan_signin_response_form=build_form("wan_signin_response_form"),
             wan_state=state_token,
             credential_options=json.dumps(o_json),
-            **_security._run_ctx_processor("wan_verify")
+            **_security._run_ctx_processor("wan_verify"),
         )
 
     if _security._want_json(request):
@@ -795,14 +816,14 @@ def webauthn_verify() -> "ResponseValue":
         wan_verify_form=form,
         wan_signin_response_form=build_form("wan_signin_response_form"),
         skip_login_menu=True,
-        **_security._run_ctx_processor("wan_verify")
+        **_security._run_ctx_processor("wan_verify"),
     )
 
 
 @auth_required(lambda: cv("API_ENABLED_METHODS"))
-def webauthn_verify_response(token: str) -> "ResponseValue":
-    form: WebAuthnSigninResponseForm = build_form_from_request(
-        "wan_signin_response_form"
+def webauthn_verify_response(token: str) -> ResponseValue:
+    form = t.cast(
+        WebAuthnSigninResponseForm, build_form_from_request("wan_signin_response_form")
     )
 
     expired, invalid, state = check_and_get_token_status(
@@ -828,7 +849,7 @@ def webauthn_verify_response(token: str) -> "ResponseValue":
         # update last use and sign count
         after_this_request(view_commit)
         assert form.cred
-        form.cred.lastuse_datetime = datetime.datetime.utcnow()
+        form.cred.lastuse_datetime = _security.datetime_factory()
         form.cred.sign_count = form.authentication_verification.new_sign_count
         _datastore.put(form.cred)
 
@@ -852,7 +873,7 @@ def webauthn_verify_response(token: str) -> "ResponseValue":
     return redirect(url_for_security("wan_verify"))
 
 
-def is_cred_usable(cred: "WebAuthn", usage: t.Union[str, t.List[str]]) -> bool:
+def is_cred_usable(cred: WebAuthn, usage: str | list[str]) -> bool:
     # Return True is cred can be used for the requested usage/verify
     if not isinstance(usage, list):
         usage = [usage]
@@ -860,7 +881,7 @@ def is_cred_usable(cred: "WebAuthn", usage: t.Union[str, t.List[str]]) -> bool:
     return cred.usage in usage
 
 
-def has_webauthn(user: "User", usage: t.Union[str, t.List[str]]) -> bool:
+def has_webauthn(user: User, usage: str | list[str]) -> bool:
     # Return True if ``user`` has one or more keys with requested usage.
     # Usage: either "first" or "secondary"
     if not isinstance(usage, list):
@@ -873,8 +894,8 @@ def has_webauthn(user: "User", usage: t.Union[str, t.List[str]]) -> bool:
 
 
 def create_credential_list(
-    user: "User", usage: t.List[str]
-) -> t.List["PublicKeyCredentialDescriptor"]:
+    user: User, usage: list[str]
+) -> list[PublicKeyCredentialDescriptor]:
     # Return a list of registered credentials - filtered by whether they apply to our
     # authentication state (first or secondary)
     cl = []
@@ -896,25 +917,25 @@ def create_credential_list(
 
 
 class WebAuthnTfPlugin(TfPluginBase):
-    def __init__(self, app: "flask.Flask"):
+    def __init__(self, app: flask.Flask):
         super().__init__(app)
 
     def create_blueprint(
-        self, app: "flask.Flask", bp: "flask.Blueprint", state: "Security"
+        self, app: flask.Flask, bp: flask.Blueprint, state: Security
     ) -> None:
         """Our endpoints are already registered since webauthn can be both
         a 'first' or 'secondary' authentication mechanism.
         """
         pass
 
-    def get_setup_methods(self, user: "User") -> t.List[t.Optional[str]]:
+    def get_setup_methods(self, user: User) -> list[str]:
         if has_webauthn(user, "secondary"):
             return [_("webauthn")]
         return []
 
     def tf_login(
-        self, user: "User", json_payload: t.Dict[str, t.Any], next_loc: t.Optional[str]
-    ) -> "ResponseValue":
+        self, user: User, json_payload: dict[str, t.Any], next_loc: str | None
+    ) -> ResponseValue:
         session["tf_state"] = "ready"
         if not _security._want_json(request):
             values = dict(next=next_loc) if next_loc else dict()
